# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.


import json
import time
from typing import Any, List, Optional, Union

from deprecated.sphinx import deprecated
from opensearchpy import OpenSearch

from opensearch_py_ml.ml_commons.ml_common_utils import (
    ML_BASE_URI,
    MODEL_FORMAT_FIELD,
    MODEL_GROUP_ID,
    MODEL_NAME_FIELD,
    MODEL_VERSION_FIELD,
    TIMEOUT,
)
from opensearch_py_ml.ml_commons.model_access_control import ModelAccessControl
from opensearch_py_ml.ml_commons.model_connector import Connector
from opensearch_py_ml.ml_commons.model_execute import ModelExecute
from opensearch_py_ml.ml_commons.model_uploader import ModelUploader
from opensearch_py_ml.ml_commons.validators import validate_profile_input


class MLCommonClient:
    """
    A client that communicates to the ml-common plugin for OpenSearch. This client allows for registration of trained
    machine learning models to an OpenSearch index.
    """

    def __init__(self, os_client: OpenSearch):
        self._client = os_client
        self._model_uploader = ModelUploader(os_client)
        self._model_execute = ModelExecute(os_client)
        self.model_access_control = ModelAccessControl(os_client)
        self.connector = Connector(os_client)

    def execute(self, algorithm_name: str, input_json: dict) -> dict:
        """
        This method executes ML algorithms that can be only executed directly (i.e. do not support train and
        predict APIs), like anomaly localization and metrics correlation. The algorithm has to be supported by ML Commons.
        Refer to https://opensearch.org/docs/2.7/ml-commons-plugin/api/#execute

        :param algorithm_name: Name of the algorithm
        :type algorithm_name: string
        :param input_json: Dictionary of parameters
        :type input_json: dict
        :return: returns the API response
        :rtype: dict
        """
        return self._model_execute._execute(algorithm_name, input_json)

    @deprecated(
        reason="Since OpenSearch 2.7.0, you can use register_model instead",
        version="2.7.0",
    )
    def upload_model(
        self,
        model_path: str,
        model_config_path: str,
        isVerbose: bool = False,
        load_model: bool = True,
        wait_until_loaded: bool = True,
    ) -> str:
        """
        This method registers the model in the opensearch cluster using ml-common plugin's api.
        First, this method creates a model id to store model metadata and then breaks the model zip file into
        multiple chunks and then upload chunks into opensearch cluster

        :param model_path: path of the zip file of the model
        :type model_path: string
        :param model_config_path: filepath of the model metadata. A json file of model metadata is expected
            Model metadata format example:
            {
                "name": "all-MiniLM-L6-v2",
                "version": 1,
                "model_format": "TORCH_SCRIPT",
                "model_config": {
                    "model_type": "bert",
                    "embedding_dimension": 384,
                    "framework_type": "sentence_transformers",
                },
            }

            refer to:
            https://opensearch.org/docs/latest/ml-commons-plugin/model-serving-framework/#upload-model-to-opensearch
        :type model_config_path: string
        :param isVerbose: if isVerbose is true method will print more messages. default False
        :type isVerbose: boolean
        :param load_model: Whether to deploy the model using uploaded model chunks
        :type load_model: bool
        :param wait_until_loaded: If load_model is true, whether to wait until the model is deployed
        :type wait_until_loaded: bool
        :return: returns the model_id so that we can use this for further operation.
        :rtype: string
        """
        model_id = self._model_uploader._register_model(
            model_path=model_path,
            model_meta_path=model_config_path,
            isVerbose=isVerbose,
        )

        # loading the model chunks from model index
        if load_model:
            self.load_model(model_id, wait_until_loaded=wait_until_loaded)

        return model_id

    def train_model(
        self, algorithm_name: str, input_json: dict, is_async: Optional[bool] = False
    ) -> dict:
        """
        This method trains an ML Model
        """

        params = {}
        if not isinstance(input_json, dict):
            input_json = json.loads(input_json)
        if is_async:
            params["async"] = "true"

        return self._client.transport.perform_request(
            method="POST",
            url=f"{ML_BASE_URI}/_train/{algorithm_name}",
            body=input_json,
            params=params,
        )

    def register_model(
        self,
        model_path: str,
        model_config_path: str,
        model_group_id: str = "",
        isVerbose: bool = False,
        deploy_model: bool = True,
        wait_until_deployed: bool = True,
    ) -> str:
        """
        This method registers the model in the opensearch cluster using ml-common plugin's api.
        First, this method creates a model id to store model metadata and then breaks the model zip file into
        multiple chunks and then upload chunks into opensearch cluster

        :param model_path: path of the zip file of the model
        :type model_path: string
        :param model_config_path: filepath of the model metadata. A json file of model metadata is expected
            Model metadata format example:
            {
                "name": "all-MiniLM-L6-v2",
                "version": 1,
                "model_format": "TORCH_SCRIPT",
                "model_config": {
                    "model_type": "bert",
                    "embedding_dimension": 384,
                    "framework_type": "sentence_transformers",
                },
            }

            refer to:
            https://opensearch.org/docs/latest/ml-commons-plugin/model-serving-framework/#upload-model-to-opensearch
        :type model_config_path: string
        :param model_group_id: Model group id
        :type model_group_id: string
        :param isVerbose: if isVerbose is true method will print more messages. default False
        :type isVerbose: boolean
        :param deploy_model: Whether to deploy the model using uploaded model chunks
        :type deploy_model: bool
        :param wait_until_deployed: If deploy_model is true, whether to wait until the model is deployed
        :type wait_until_deployed: bool
        :return: returns the model_id so that we can use this for further operation.
        :rtype: string
        """
        model_id = self._model_uploader._register_model(
            model_path, model_config_path, model_group_id, isVerbose
        )

        # loading the model chunks from model index
        if deploy_model:
            self.deploy_model(model_id, wait_until_deployed=wait_until_deployed)

        return model_id

    @deprecated(
        reason="Since OpenSearch 2.7.0, you can use register_pretrained_model instead",
        version="2.7.0",
    )
    def upload_pretrained_model(
        self,
        model_name: str,
        model_version: str,
        model_format: str,
        load_model: bool = True,
        wait_until_loaded: bool = True,
    ):
        """
        This method registers the pretrained model in the opensearch cluster using ml-common plugin's api.
        First, this method creates a model id to store model info and then deploys the model if load_model is True.
        The model has to be supported by ML Commons. Refer to https://opensearch.org/docs/latest/ml-commons-plugin/pretrained-models/.

        :param model_name: Name of the pretrained model
        :type model_name: string
        :param model_version: Version of the pretrained model
        :type model_version: string
        :param model_format: "TORCH_SCRIPT" or "ONNX"
        :type model_format: string
        :param load_model: Whether to deploy the model using uploaded model chunks
        :type load_model: bool
        :param wait_until_loaded: If load_model is true, whether to wait until the model is deployed
        :type wait_until_loaded: bool
        :return: returns the model_id so that we can use this for further operation
        :rtype: string
        """
        # creating model meta doc
        model_config_json = {
            MODEL_NAME_FIELD: model_name,
            "version": model_version,
            "model_format": model_format,
        }
        model_id = self._send_model_info(model_config_json)

        # loading the model chunks from model index
        if load_model:
            self.load_model(model_id, wait_until_loaded=wait_until_loaded)

        return model_id

    def register_pretrained_model(
        self,
        model_name: str,
        model_version: str,
        model_format: str,
        model_group_id: str = "",
        deploy_model: bool = True,
        wait_until_deployed: bool = True,
    ):
        """
        This method registers the pretrained model in the opensearch cluster using ml-common plugin's api.
        First, this method creates a model id to store model info and then deploys the model if deploy_model is True.
        The model has to be supported by ML Commons. Refer to https://opensearch.org/docs/latest/ml-commons-plugin/pretrained-models/.

        :param model_name: Name of the pretrained model
        :type model_name: string
        :param model_version: Version of the pretrained model
        :type model_version: string
        :param model_format: "TORCH_SCRIPT" or "ONNX"
        :type model_format: string
        :param model_group_id: Model group id
        :type model_group_id: string
        :param deploy_model: Whether to deploy the model using uploaded model chunks
        :type deploy_model: bool
        :param wait_until_deployed: If deploy_model is true, whether to wait until the model is deployed
        :type wait_until_deployed: bool
        :return: returns the model_id so that we can use this for further operation
        :rtype: string
        """
        # creating model meta doc
        model_config_json = {
            MODEL_NAME_FIELD: model_name,
            MODEL_VERSION_FIELD: model_version,
            MODEL_FORMAT_FIELD: model_format,
            MODEL_GROUP_ID: model_group_id,
        }
        model_id = self._send_model_info(model_config_json)

        print(model_id)

        # loading the model chunks from model index
        if deploy_model:
            self.deploy_model(model_id, wait_until_deployed=wait_until_deployed)

        return model_id

    def _send_model_info(self, model_meta_json: dict):
        """
        This method sends the pretrained model info to ML Commons' register api

        :param model_meta_json: a dictionary object with model configurations
        :type model_meta_json: dict
        :return: returns a unique id of the model
        :rtype: string
        """
        output: Union[bool, Any] = self._client.transport.perform_request(
            method="POST",
            url=f"{ML_BASE_URI}/models/_register",
            body=model_meta_json,
        )
        end = time.time() + TIMEOUT  # timeout seconds
        task_flag = False
        while not task_flag and time.time() < end:
            time.sleep(1)
            status = self._get_task_info(output["task_id"])
            if status["state"] != "CREATED":
                task_flag = True
        # TODO: need to add the test case later for this line
        if not task_flag:
            raise TimeoutError("Model registration timed out")
        if status["state"] == "FAILED":
            raise Exception(status["error"])
        print("Model was registered successfully. Model Id: ", status["model_id"])
        return status["model_id"]

    @deprecated(
        reason="Since OpenSearch 2.7.0, you can use deploy_model instead",
        version="2.7.0",
    )
    def load_model(self, model_id: str, wait_until_loaded: bool = True) -> object:
        """
        This method deploys a model in the opensearch cluster using ml-common plugin's deploy model api

        :param model_id: unique id of the model
        :type model_id: string
        :param wait_until_loaded: Whether to wait until the model is deployed
        :type wait_until_loaded: bool
        :return: returns a json object, with task_id and status key.
        :rtype: object
        """

        API_URL = f"{ML_BASE_URI}/models/{model_id}/_deploy"

        task_id = self._client.transport.perform_request(method="POST", url=API_URL)[
            "task_id"
        ]

        if wait_until_loaded:
            # Wait until deployed
            for i in range(TIMEOUT):
                ml_model_status = self.get_model_info(model_id)
                model_state = ml_model_status.get("model_state")
                if model_state in ["DEPLOYED", "PARTIALLY_DEPLOYED"]:
                    break
                time.sleep(1)

            # TODO: need to add the test case later for this line
            # Check the model status
            if model_state == "DEPLOYED":
                print("Model deployed successfully")
            elif model_state == "PARTIALLY_DEPLOYED":
                print("Model deployed only partially")
            # TODO: need to add the test case later for this line
            else:
                raise Exception("Model deployment failed")

        return self._get_task_info(task_id)

    def deploy_model(self, model_id: str, wait_until_deployed: bool = True) -> object:
        """
        This method deploys a model in the opensearch cluster using ml-common plugin's deploy model api

        :param model_id: unique id of the model
        :type model_id: string
        :param wait_until_deployed: Whether to wait until the model is deployed
        :type wait_until_deployed: bool
        :return: returns a json object, with task_id and status key.
        :rtype: object
        """

        API_URL = f"{ML_BASE_URI}/models/{model_id}/_deploy"

        task_id = self._client.transport.perform_request(method="POST", url=API_URL)[
            "task_id"
        ]

        print(f"Task ID: {task_id}")
        if wait_until_deployed:
            # Wait until deployed
            for i in range(TIMEOUT):
                ml_model_status = self.get_model_info(model_id)
                model_state = ml_model_status.get("model_state")
                if model_state in ["DEPLOYED", "PARTIALLY_DEPLOYED"]:
                    break
                time.sleep(1)

            # TODO: need to add the test case later for this line
            # Check the model status
            if model_state == "DEPLOYED":
                print("Model deployed successfully")
            elif model_state == "PARTIALLY_DEPLOYED":
                print("Model deployed only partially")
            else:
                raise Exception("Model deployment failed")

        return self._get_task_info(task_id)

    def get_task_info(self, task_id: str, wait_until_task_done: bool = False) -> object:
        """
        This method return information about a task running into opensearch cluster (using ml commons api)
        when we deploy a model

        :param task_id: unique id of the task
        :type task_id: string
        :param wait_until_task_done: a boolean indicator if we want to wait until a task done before
            returning the task related information
        :type wait_until_task_done: bool
        :return: returns a json object, with detailed information about the task
        :rtype: object
        """
        if wait_until_task_done:
            end = time.time() + TIMEOUT  # timeout seconds
            task_flag = False
            while not task_flag and time.time() < end:
                time.sleep(1)
                output = self._get_task_info(task_id)
                if (
                    output["state"] == "COMPLETED"
                    or output["state"] == "FAILED"
                    or output["state"] == "COMPLETED_WITH_ERROR"
                ):
                    task_flag = True
        return self._get_task_info(task_id)

    def _get_task_info(self, task_id: str):
        API_URL = f"{ML_BASE_URI}/tasks/{task_id}"

        return self._client.transport.perform_request(
            method="GET",
            url=API_URL,
        )

    def search_task(self, input_json) -> object:
        """
        This method searches a task from opensearch cluster (using ml commons api)
        :param json: json input for the search request
        :type json: string or dict
        :return: returns a json object, with detailed information about the searched task
        :rtype: object
        """

        API_URL = f"{ML_BASE_URI}/tasks/_search"

        if isinstance(input_json, str):
            try:
                json_obj = json.loads(input_json)
                if not isinstance(json_obj, dict):
                    return "Invalid JSON object passed as argument."
                API_BODY = json.dumps(json_obj)
            except json.JSONDecodeError:
                return "Invalid JSON string passed as argument."
        elif isinstance(input_json, dict):
            API_BODY = json.dumps(input_json)
        else:
            return "Invalid JSON object passed as argument."

        return self._client.transport.perform_request(
            method="GET",
            url=API_URL,
            body=API_BODY,
        )

    def search_model(self, input_json) -> object:
        """
        This method searches a model from opensearch cluster (using ml commons api)
        :param json: json input for the search request
        :type json: string or dict
        :return: returns a json object, with detailed information about the searched model
        :rtype: object
        """

        API_URL = f"{ML_BASE_URI}/models/_search"

        if isinstance(input_json, str):
            try:
                json_obj = json.loads(input_json)
                if not isinstance(json_obj, dict):
                    return "Invalid JSON object passed as argument."
                API_BODY = json.dumps(json_obj)
            except json.JSONDecodeError:
                return "Invalid JSON string passed as argument."
        elif isinstance(input_json, dict):
            API_BODY = json.dumps(input_json)
        else:
            return "Invalid JSON object passed as argument."

        return self._client.transport.perform_request(
            method="POST",
            url=API_URL,
            body=API_BODY,
        )

    def get_model_info(self, model_id: str) -> object:
        """
        This method return information about a model registered in the opensearch cluster (using ml commons api)

        :param model_id: unique id of the model
        :type model_id: string
        :return: returns a json object, with detailed information about the model
        :rtype: object
        """

        API_URL = f"{ML_BASE_URI}/models/{model_id}"

        return self._client.transport.perform_request(
            method="GET",
            url=API_URL,
        )

    def generate_model_inference(self, model_id: str, request_body: dict) -> object:
        """
        Generates inference result for the given input using the specified request body.

        :param model_id: Unique ID of the model.
        :type model_id: string
        :param request_body: Request body to send to the API.
        :type request_body: dict
        :return: Returns a JSON object `inference_results` containing the results for the given input.
        :rtype: object
        """
        API_URL = f"{ML_BASE_URI}/models/{model_id}/_predict/"
        return self._client.transport.perform_request(
            method="POST",
            url=API_URL,
            body=request_body,
        )

    def generate_embedding(self, model_id: str, sentences: List[str]) -> object:
        """
        This method return embedding for given sentences (using ml commons _predict api)

        :param model_id: unique id of the nlp model
        :type model_id: string
        :param sentences: List of sentences
        :type sentences: list of string
        :return: returns a json object `inference_results` which is a list of embedding results of given sentences
            every item has 4 properties: name, data_type, shape, data (embedding value)
        :rtype: object
        """

        API_URL = f"{ML_BASE_URI}/_predict/text_embedding/{model_id}"

        API_BODY = {"text_docs": sentences, "target_response": ["sentence_embedding"]}

        return self._client.transport.perform_request(
            method="POST",
            url=API_URL,
            body=API_BODY,
        )

    def generate_question_answering(
        self, model_id: str, question: str, context: str
    ) -> object:
        """
        This method return answer for given question and context (using ml commons _predict api)

        :param model_id: unique id of the question answering model
        :type model_id: string
        :param question: question
        :type question: string
        :param context: context
        :type context: string
        :return: returns a json object `inference_results` which is a list of answer results of given question and context
        :rtype: object
        """
        API_URL = f"{ML_BASE_URI}/_predict/question_answering/{model_id}"
        API_BODY = {"question": question, "context": context}
        return self._client.transport.perform_request(
            method="POST",
            url=API_URL,
            body=API_BODY,
        )

    @deprecated(
        reason="Since OpenSearch 2.7.0, you can use undeploy_model instead",
        version="2.7.0",
    )
    def unload_model(self, model_id: str, node_ids: List[str] = []) -> object:
        """
        This method undeploys a model from all the nodes or from the given list of nodes (using ml commons _undeploy api)

        :param model_id: unique id of the nlp model
        :type model_id: string
        :param node_ids: List of nodes
        :type node_ids: list of string
        :return: returns a json object with defining from which nodes the model was undeployed.
        :rtype: object
        """

        API_URL = f"{ML_BASE_URI}/models/{model_id}/_undeploy"

        # TODO: need to add the test case later for this line
        API_BODY = {}
        if len(node_ids) > 0:
            API_BODY["node_ids"] = node_ids

        return self._client.transport.perform_request(
            method="POST",
            url=API_URL,
            body=API_BODY,
        )

    def predict(
        self, model_id: str, predict_object: dict, algorithm_name: str = None
    ) -> dict:
        """
        Generalized predict method to make predictions using different ML algorithms.

        :param algorithm_name: The name of the algorithm, e.g., 'kmeans', 'text_embedding'
        :type algorithm_name: str
        :param model_id: Unique identifier of the deployed model
        :type model_id: str
        :param predict_object: JSON object containing the input data and parameters for prediction
        :type predict_object: dict
        :return: Prediction response from the ML model
        :rtype: dict
        """
        # Make the POST request to the prediction API
        response = self.generate_model_inference(model_id, predict_object)

        return response

    def undeploy_model(self, model_id: str, node_ids: List[str] = []) -> object:
        """
        This method undeploys a model from all the nodes or from the given list of nodes (using ml commons _undeploy api)

        :param model_id: unique id of the nlp model
        :type model_id: string
        :param node_ids: List of nodes
        :type node_ids: list of string
        :return: returns a json object with defining from which nodes the model was undeployed.
        :rtype: object
        """

        API_URL = f"{ML_BASE_URI}/models/{model_id}/_undeploy"

        API_BODY = {}
        if len(node_ids) > 0:
            API_BODY["node_ids"] = node_ids

        return self._client.transport.perform_request(
            method="POST",
            url=API_URL,
            body=API_BODY,
        )

    def delete_model(self, model_id: str) -> object:
        """
        This method deletes a model from opensearch cluster (using ml commons api)

        :param model_id: unique id of the model
        :type model_id: string
        :return: returns a json object, with detailed information about the deleted model
        :rtype: object
        """

        API_URL = f"{ML_BASE_URI}/models/{model_id}"

        return self._client.transport.perform_request(
            method="DELETE",
            url=API_URL,
        )

    def delete_task(self, task_id: str) -> object:
        """
        This method deletes a task from opensearch cluster (using ml commons api)

        :param task_id: unique id of the task
        :type task_id: string
        :return: returns a json object, with detailed information about the deleted task
        :rtype: object
        """

        API_URL = f"{ML_BASE_URI}/tasks/{task_id}"

        return self._client.transport.perform_request(
            method="DELETE",
            url=API_URL,
        )

    def _get_profile(self, payload: Optional[dict] = None):
        """
        Get the profile using the given payload.

        :param payload: The payload to be used for getting the profile. Defaults to None.
        :type payload: Optional[dict]
        :return: The response from the server after performing the request.
        :rtype: Any
        """
        validate_profile_input(None, payload)
        return self._client.transport.perform_request(
            method="GET", url=f"{ML_BASE_URI}/profile", body=payload
        )

    def _get_models_profile(
        self, model_id: Optional[str] = "", payload: Optional[dict] = None
    ):
        """
        Get the profile of a model.

        Args:
            model_id (str, optional): The ID of the model. Defaults to "".
            payload (dict, optional): Additional payload for the request. Defaults to None.

        Returns:
            dict: The response from the API.
        """
        validate_profile_input(model_id, payload)

        url = f"{ML_BASE_URI}/profile/models/{model_id if model_id else ''}"
        return self._client.transport.perform_request(
            method="GET", url=url, body=payload
        )

    def _get_tasks_profile(
        self, task_id: Optional[str] = "", payload: Optional[dict] = None
    ):
        """
        Retrieves the profile of a task from the API.

        Parameters:
            task_id (str, optional): The ID of the task to retrieve the profile for. Defaults to an empty string.
            payload (dict, optional): Additional payload for the request. Defaults to None.

        Returns:
            dict: The profile of the task.

        Raises:
            ValueError: If the input validation fails.

        """
        validate_profile_input(task_id, payload)

        url = f"{ML_BASE_URI}/profile/tasks/{task_id if task_id else ''}"
        return self._client.transport.perform_request(
            method="GET", url=url, body=payload
        )

    def get_profile(
        self,
        profile_type: str = "all",
        ids: Optional[Union[str, List[str]]] = None,
        request_body: Optional[dict] = None,
    ) -> dict:
        """
        Get profile information based on the profile type.

        Args:
            profile_type: The type of profile to retrieve. Valid values are 'all', 'model', or 'task'. Default is 'all'.
                'all': Retrieves all profiles available.
                'model': Retrieves the profile(s) of the specified model(s). The model(s) to retrieve are specified by the 'ids' parameter.
                'task': Retrieves the profile(s) of the specified task(s). The task(s) to retrieve are specified by the 'ids' parameter.
            ids: Either a single profile ID as a string, or a list of profile IDs to retrieve. Default is None.
            request_body: The request body containing additional information. Default is None.

        Returns:
            The profile information.

        Raises:
            ValueError: If the profile_type is not 'all', 'model', or 'task'.

        Example:
            get_profile()

            get_profile(profile_type='model', ids='model1')

            get_profile(profile_type='model', ids=['model1', 'model2'])

            get_profile(profile_type='task', ids='task1', request_body={"node_ids": ["KzONM8c8T4Od-NoUANQNGg"],"return_all_tasks": true,"return_all_models": true})

            get_profile(profile_type='task', ids=['task1', 'task2'], request_body={'additional': 'info'})
        """

        if profile_type == "all":
            return self._get_profile(request_body)
        elif profile_type == "model":
            if ids and isinstance(ids, list):
                ids = ",".join(ids)
            return self._get_models_profile(ids, request_body)
        elif profile_type == "task":
            if ids and isinstance(ids, list):
                ids = ",".join(ids)
            return self._get_tasks_profile(ids, request_body)
        else:
            raise ValueError(
                "Invalid profile type. Profile type must be 'all', 'model' or 'task'."
            )



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

BANDWIDTH = 16.0
NMF_MAX_ITER = 300
NMF_TOL = 1e-4
DT_PVAL = 0.5
OMP_TOL = 0.8



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.


#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

# TODO fill out this information
__title__ = "opensearch_py_ml"
__description__ = "Python Client and Toolkit for DataFrames, Big Data, Machine Learning and ETL in OpenSearch"
__url__ = "https://github.com/opensearch-project/opensearch-py-ml"
__version__ = "1.3.0"
__author__ = "Dhrubo Saha"
__author_email__ = "dhrubo@amazon.com, "
__maintainer__ = "Dhrubo Saha, Yaliang Wu, Alexander Greaves-Tunnell, Sicheng Song, Jing Zhang, Bhavana Goud Ramaram, Yerzhaisang Taskali, Nathalie Jonathan"
__maintainer_email__ = "dhrubo@amazon.com, ylwu@amazon.com, greaa@amazon.com, seasonsg@amazon.com, jngz@amazon.com, rbhavna@amazon.com, tasqali1697@gmail.com, nathhjo@amazon.com"



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

import json
import os
from math import ceil
from typing import Any, Iterable, Union

from opensearchpy import OpenSearch

from opensearch_py_ml.ml_commons.ml_common_utils import (
    EMBEDDING_DIMENSION,
    FRAMEWORK_TYPE,
    META_API_ENDPOINT,
    ML_BASE_URI,
    MODEL_CHUNK_MAX_SIZE,
    MODEL_CONFIG_FIELD,
    MODEL_CONTENT_HASH_VALUE,
    MODEL_CONTENT_SIZE_IN_BYTES_FIELD,
    MODEL_FORMAT_FIELD,
    MODEL_FUNCTION_NAME,
    MODEL_GROUP_ID,
    MODEL_MAX_SIZE,
    MODEL_NAME_FIELD,
    MODEL_TYPE,
    MODEL_VERSION_FIELD,
    TOTAL_CHUNKS_FIELD,
    _generate_model_content_hash_value,
)


class ModelUploader:
    """
    Class for registering a model using ml-commons apis in opensearch cluster.
    """

    def __init__(self, os_client: OpenSearch):
        self._client = os_client

    def _register_model(
        self,
        model_path: str,
        model_meta_path: str,
        model_group_id: str = "",
        isVerbose: bool = False,
    ) -> str:
        """
        This method registers the model in the opensearch cluster using ml-common plugin's register model api.
        First, this method creates a model id to store model metadata and then breaks the model zip file into
        multiple chunks and then upload chunks into cluster.

        :param model_path: path of the zip file of the model
        :type model_path: string
        :param model_meta_path:
            filepath of the model metadata. A json file of model metadata is expected
            Model metadata format example:
                {
                    "name": "all-MiniLM-L6-v2",
                    "version": 1,
                    "model_format": "TORCH_SCRIPT",
                    "model_config": {
                        "model_type": "bert",
                        "embedding_dimension": 384,
                        "framework_type": "sentence_transformers",
                        "all_config": '{"_name_or_path":"nreimers/MiniLM-L6-H384-uncased","architectures":["BertModel"],"attention_probs_dropout_prob":0.1,"gradient_checkpointing":false,"hidden_act":"gelu","hidden_dropout_prob":0.1,"hidden_size":384,"initializer_range":0.02,"intermediate_size":1536,"layer_norm_eps":1e-12,"max_position_embeddings":512,"model_type":"bert","num_attention_heads":12,"num_hidden_layers":6,"pad_token_id":0,"position_embedding_type":"absolute","transformers_version":"4.8.2","type_vocab_size":2,"use_cache":true,"vocab_size":30522}',
                    },
                }
            refer to:
                https://opensearch.org/docs/latest/ml-commons-plugin/model-serving-framework/#upload-model-to-opensearch
        :type model_meta_path: string
        :param model_group_id: Model group id
        :type model_group_id: string
        :param isVerbose: if isVerbose is true method will print more messages
        :type isVerbose: bool
        :return: returns model id which is created by the model metadata
        :rtype: string
        """
        if os.stat(model_path).st_size > MODEL_MAX_SIZE:
            raise Exception("Model file size exceeds the limit of 4GB")

        model_content_size_in_bytes = os.stat(model_path).st_size
        total_num_chunks: int = ceil(model_content_size_in_bytes / MODEL_CHUNK_MAX_SIZE)

        if isVerbose:
            print("Total number of chunks", total_num_chunks)

        model_meta_json_file = open(model_meta_path)

        model_meta_json: dict[str, Union[str, dict[str, str]]] = json.load(
            model_meta_json_file
        )
        model_meta_json[TOTAL_CHUNKS_FIELD] = total_num_chunks

        if MODEL_CONTENT_SIZE_IN_BYTES_FIELD not in model_meta_json:
            model_meta_json[MODEL_CONTENT_SIZE_IN_BYTES_FIELD] = (
                model_content_size_in_bytes
            )
        if MODEL_CONTENT_HASH_VALUE not in model_meta_json:
            # Generate the sha1 hash for the model zip file
            hash_val_model_file = _generate_model_content_hash_value(model_path)
            model_meta_json[MODEL_CONTENT_HASH_VALUE] = hash_val_model_file
            if isVerbose:
                print("Sha1 value of the model file: ", hash_val_model_file)

        model_meta_json[MODEL_GROUP_ID] = model_group_id

        if self._check_mandatory_field(model_meta_json):
            meta_output: Union[bool, Any] = self._client.transport.perform_request(
                method="POST",
                url=f"{ML_BASE_URI}/{META_API_ENDPOINT}",
                body=model_meta_json,
            )
            print(
                "Model meta data was created successfully. Model Id: ",
                meta_output.get("model_id"),
            )

            # model meta doc is created successfully, and now we can upload model chunks to the model id
            if meta_output.get("status") == "CREATED" and meta_output.get("model_id"):
                model_id = meta_output.get("model_id")

                def model_file_chunk_generator() -> Iterable[str]:
                    with open(model_path, "rb") as f:
                        while True:
                            data = f.read(MODEL_CHUNK_MAX_SIZE)
                            if not data:
                                break
                            yield data  # type: ignore # check if we actually need to do base64 encoding

                to_iterate_over = enumerate(model_file_chunk_generator())

                for i, chunk in to_iterate_over:
                    if isVerbose:
                        print(f"uploading chunk {i + 1} of {total_num_chunks}")
                    output = self._client.transport.perform_request(
                        method="POST",
                        url=f"{ML_BASE_URI}/models/{model_id}/chunk/{i}",
                        body=chunk,
                    )
                    if isVerbose:
                        print("Model id:", output)

                print("Model registered successfully")
                return model_id
            else:
                raise Exception(
                    "Model meta doc creation wasn't successful. Please check the errors"
                )

    def _check_mandatory_field(self, model_meta: dict) -> bool:
        """
        This method checks if model meta doc has all the required fields to create a model meta doc in opensearch

        Parameters
        ----------
        :param model_meta: content of the model meta file
        :type model_meta: dict

        Returns
        -------
        :return: if all the required fields are present returns True otherwise
                                            raise exception
        :rtype: bool
        """

        if model_meta:

            if not model_meta.get(MODEL_NAME_FIELD):
                raise ValueError(f"{MODEL_NAME_FIELD} can not be empty")
            if not model_meta.get(MODEL_VERSION_FIELD):
                raise ValueError(f"{MODEL_VERSION_FIELD} can not be empty")
            if not model_meta.get(MODEL_FORMAT_FIELD):
                raise ValueError(f"{MODEL_FORMAT_FIELD} can not be empty")
            if not model_meta.get(MODEL_CONTENT_HASH_VALUE):
                raise ValueError(f"{MODEL_CONTENT_HASH_VALUE} can not be empty")
            if not model_meta.get(TOTAL_CHUNKS_FIELD):
                raise ValueError(f"{TOTAL_CHUNKS_FIELD} can not be empty")
            if not model_meta.get(MODEL_CONFIG_FIELD):
                if model_meta.get(MODEL_FUNCTION_NAME) not in [
                    "SPARSE_ENCODING",
                    "SPARSE_TOKENIZE",
                ]:
                    raise ValueError(f"{MODEL_CONFIG_FIELD} can not be empty")
            else:
                if not isinstance(model_meta.get(MODEL_CONFIG_FIELD), dict):
                    raise TypeError(
                        f"{MODEL_CONFIG_FIELD} is expecting to be an object"
                    )
                model_config = model_meta.get(MODEL_CONFIG_FIELD)
                if not model_config.get(MODEL_TYPE):
                    raise ValueError(f"{MODEL_TYPE} can not be empty")
                if model_config.get(MODEL_TYPE) != "sentence_highlighting":
                    if not model_config.get(EMBEDDING_DIMENSION):
                        raise ValueError(f"{EMBEDDING_DIMENSION} can not be empty")
                if not model_config.get(FRAMEWORK_TYPE):
                    raise ValueError(f"{FRAMEWORK_TYPE} can not be empty")
            return True
        else:
            raise ValueError("Model metadata can't be empty")



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.
import json
import os
from abc import ABC, abstractmethod
from zipfile import ZipFile

import requests

from opensearch_py_ml.ml_commons.ml_common_utils import (
    LICENSE_URL,
    SPARSE_ENCODING_FUNCTION_NAME,
)


class BaseUploadModel(ABC):
    """
    A base class for uploading models to OpenSearch pretrained model hub.
    """

    def __init__(
        self, model_id: str, folder_path: str = None, overwrite: bool = False
    ) -> None:
        self.model_id = model_id
        self.folder_path = folder_path
        self.overwrite = overwrite

    @abstractmethod
    def save_as_pt(self, *args, **kwargs):
        pass

    @abstractmethod
    def save_as_onnx(self, *args, **kwargs):
        pass

    @abstractmethod
    def make_model_config_json(
        self,
        version_number: str,
        model_format: str,
        description: str,
    ) -> str:
        pass

    def _fill_null_truncation_field(
        self,
        save_json_folder_path: str,
        max_length: int,
    ) -> None:
        """
        Fill truncation field in tokenizer.json when it is null

        :param save_json_folder_path:
             path to save model json file, e.g, "home/save_pre_trained_model_json/")
        :type save_json_folder_path: string
        :param max_length:
             maximum sequence length for model
        :type max_length: int
        :return: no return value expected
        :rtype: None
        """
        tokenizer_file_path = os.path.join(save_json_folder_path, "tokenizer.json")
        with open(tokenizer_file_path) as user_file:
            parsed_json = json.load(user_file)
        if "truncation" not in parsed_json or parsed_json["truncation"] is None:
            parsed_json["truncation"] = {
                "direction": "Right",
                "max_length": max_length,
                "strategy": "LongestFirst",
                "stride": 0,
            }
            with open(tokenizer_file_path, "w") as file:
                json.dump(parsed_json, file, indent=2)

    def _add_apache_license_to_model_zip_file(self, model_zip_file_path: str):
        """
        Add Apache-2.0 license file to the model zip file at model_zip_file_path

        :param model_zip_file_path:
            Path to the model zip file
        :type model_zip_file_path: string
        :return: no return value expected
        :rtype: None
        """
        r = requests.get(LICENSE_URL)
        assert r.status_code == 200, "Failed to add license file to the model zip file"

        with ZipFile(str(model_zip_file_path), "a") as zipObj:
            zipObj.writestr("LICENSE", r.content)


class SparseModel(BaseUploadModel, ABC):
    """
    Class for autotracing the Sparse Encoding model.
    """

    def __init__(
        self,
        model_id: str,
        folder_path: str = "./model_files/",
        overwrite: bool = False,
    ):
        super().__init__(model_id, folder_path, overwrite)
        self.model_id = model_id
        self.folder_path = folder_path
        self.overwrite = overwrite
        self.function_name = SPARSE_ENCODING_FUNCTION_NAME

    def pre_process(self):
        pass

    def post_process(self):
        pass



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

import json
import logging
import os
from typing import Any, Dict, Optional, Tuple, Union
from urllib.parse import urlparse

import yaml
from colorama import Fore, Style
from rich.console import Console

from opensearch_py_ml.ml_commons.cli.ai_connector_helper import AIConnectorHelper
from opensearch_py_ml.ml_commons.cli.aws_config import AWSConfig
from opensearch_py_ml.ml_commons.cli.opensearch_domain_config import (
    OpenSearchDomainConfig,
)

# Import readline to enable arrow key navigation, import it this way to avoid lint error
__import__("readline")

# Initialize Rich console for enhanced CLI outputs
console = Console()

# Configure the logger for this module
logger = logging.getLogger(__name__)


class CLIBase:

    # Default setup configuration and output file name
    CONFIG_FILE = os.path.join(os.getcwd(), "setup_config.yml")
    OUTPUT_FILE = os.path.join(os.getcwd(), "output.yml")

    OPEN_SOURCE = AIConnectorHelper.OPEN_SOURCE
    AMAZON_OPENSEARCH_SERVICE = AIConnectorHelper.AMAZON_OPENSEARCH_SERVICE

    def __init__(self):
        """
        Initialize the CLIBase class.
        """
        self.config = {}
        self.output_config = {
            "connector_create": [],
            "register_model": [],
            "predict_model": [],
        }

    def load_config(
        self, config_path: str = None, config_type: str = "setup"
    ) -> Dict[str, Any]:
        """
        Load configuration from the specified file path.

        Args:
            config_path (optional): Path to the configuration file.
            config_type (optional): Type of configuration being loaded. Defaults to "setup".

        Returns:
            Dict[str, Any]: The loaded configuration as a dictionary. Returns an empty dictionary if:
                - The file doesn't exist
                - There are permission issues
                - The YAML is invalid
                - The file is empty
                - Any other error occurs
        """
        try:
            # Handle config path for setup configuration
            if config_type == "setup":
                config_path = config_path or self.CONFIG_FILE

            # Normalize the path
            config_path = os.path.abspath(os.path.expanduser(config_path))

            # Check if file exists
            if not os.path.exists(config_path):
                logger.warning(
                    f"{Fore.YELLOW}Configuration file not found at {config_path}{Style.RESET_ALL}"
                )
                return {}

            with open(config_path, "r") as file:
                config = yaml.safe_load(file) or {}

                # Update stored config for setup configuration
                if config_type == "setup":
                    self.CONFIG_FILE = config_path
                    self.config = config

                print(
                    f"{Fore.GREEN}\n{config_type.capitalize()} configuration loaded successfully from {config_path}{Style.RESET_ALL}"
                )
                return config
        except yaml.YAMLError as ye:
            logger.error(
                f"{Fore.RED}Error parsing YAML configuration: {str(ye)}{Style.RESET_ALL}"
            )
        except PermissionError:
            logger.error(
                f"{Fore.RED}Permission denied: Unable to read {config_path}{Style.RESET_ALL}"
            )
        except Exception as e:
            logger.error(
                f"{Fore.RED}Error loading {config_type} configuration: {str(e)}{Style.RESET_ALL}"
            )
        return {}

    def save_yaml_file(
        self,
        config: Dict[str, Any],
        file_type: str = "configuration",
        merge_existing: bool = False,
    ) -> Optional[str]:
        """
        Save data to a YAML file with optional merging of existing content.

        Args:
            config: The configuration dictionary to save.
            file_type (optional): Type of file being saved. Defaults to "configuration". Common values:
                - "configuration": Uses CONFIG_FILE as default path
                - "output": Uses OUTPUT_FILE as default path
            merge_existing (optional): Whether to merge with existing file content
                if the file exists. Defaults to False. If False and file exists, will
                prompt for overwrite confirmation.

        Returns:
            Optional[str]: The absolute path where the file was saved, or None if:
                - User cancelled the operation
                - Permission error occurred
                - Any other error during saving
                - User declined to overwrite existing file
        """
        try:
            # Determine default path and prompt message based on file type
            default_path = (
                self.CONFIG_FILE if file_type == "configuration" else self.OUTPUT_FILE
            )

            # Check write permissions before attempting any operations
            directory = os.path.dirname(default_path) or os.getcwd()
            if not os.access(directory, os.W_OK):
                raise PermissionError(f"No write access to directory: {directory}")

            # Get save path from user
            path = (
                input(
                    f"\nEnter the path to save the {file_type} information, "
                    f"or press Enter to save it in the current directory [{default_path}]: "
                ).strip()
                or default_path
            )

            # Validate and normalize the path
            path = os.path.abspath(os.path.expanduser(path))
            if not path.endswith((".yaml", ".yml")):
                path = f"{path}.yaml"

            if merge_existing and os.path.exists(path):
                # Read and merge with existing content
                config = self._merge_configs(path, config)
            elif os.path.exists(path):
                # Ask for confirmation to overwrite
                if not self._confirm_overwrite(path):
                    return None

            # Create directory if it doesn't exist
            directory = os.path.dirname(path)
            if directory and not os.path.exists(directory):
                os.makedirs(directory)

            # Save file
            with open(path, "w") as file:
                yaml.dump(config, file, default_flow_style=False, sort_keys=False)

            # Update relevant path attribute
            if file_type == "configuration":
                self.CONFIG_FILE = path
            else:
                self.OUTPUT_FILE = path

            print(
                f"{Fore.GREEN}\n{file_type.capitalize()} information saved successfully to {path}{Style.RESET_ALL}"
            )
            return path

        except PermissionError as pe:
            logger.error(f"{Fore.RED}Permission denied: {str(pe)}{Style.RESET_ALL}")
        except KeyboardInterrupt:
            logger.error(
                f"\n{Fore.YELLOW}Operation cancelled by user.{Style.RESET_ALL}"
            )
        except Exception as e:
            logger.error(
                f"{Fore.RED}Error saving {file_type}: {str(e)}{Style.RESET_ALL}"
            )
        return None

    def _merge_configs(self, path: str, new_config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Read and merge the config content.
        """
        # Read existing content
        with open(path, "r") as file:
            existing_config = yaml.safe_load(file) or {}

        # Merge new config with existing config
        for key, value in new_config.items():
            if key not in existing_config:
                existing_config[key] = value
            else:
                existing_config[key].extend(value)
        return existing_config

    def _confirm_overwrite(self, path: str) -> bool:
        """
        Ask for confirmation to overwrite existing file.
        """
        while True:
            response = (
                input(
                    f"{Fore.YELLOW}File already exists at {path}. "
                    f"Do you want to overwrite it? (yes/no): {Style.RESET_ALL}"
                )
                .strip()
                .lower()
            )
            if response in ["yes", "no"]:
                break
            logger.warning(f"{Fore.YELLOW}Please enter 'yes' or 'no'.{Style.RESET_ALL}")

        if response == "no":
            logger.warning(
                f"{Fore.YELLOW}Operation cancelled. Please choose a different path.{Style.RESET_ALL}"
            )
            return False
        return True

    def update_config(self, config: Dict[str, Any], config_path: str) -> bool:
        """
        Update config file with new configurations.

        Args:
            config:  The configuration dictionary to save.
            config_path: The file path where the configuration should be saved.

        Returns:
            bool: True if the configuration was successfully saved, False if any error
                occurred during the save operation.
        """
        try:
            with open(config_path, "w") as file:
                yaml.dump(config, file, default_flow_style=False, sort_keys=False)
            print(
                f"{Fore.GREEN}Configuration saved successfully to {config_path}.{Style.RESET_ALL}"
            )
            return True
        except Exception as e:
            logger.error(
                f"{Fore.RED}Error saving configuration: {str(e)}{Style.RESET_ALL}"
            )
            return False

    def connector_output(
        self,
        output_id: str,
        output_config: str,
        role_name: Optional[str] = None,
        role_arn: Optional[str] = None,
        secret_name: Optional[str] = None,
        secret_arn: Optional[str] = None,
    ) -> None:
        """
        Save connector output to a YAML file.

        Args:
            output_id: The connector ID created.
            output_config: The connector configuration output.
            role_name (optional): Name of the IAM role associated with the connector. Defaults to None.
            role_arn (optional): IAM role ARN associated with the connector. Defaults to None.
            secret_name (optional): Name of the secret associated with the connector. Defaults to None.
            secret_arn (optional): Secret ARN associated with the connector. Defaults to None.
        """
        connector_data = json.loads(output_config)
        connector_name = connector_data.get("name")
        # Update the connector_create section
        self.output_config["connector_create"].append(
            {
                "connector_id": output_id,
                "connector_name": connector_name,
                "connector_role_arn": role_arn or "",
                "connector_role_name": role_name or "",
                "connector_secret_arn": secret_arn or "",
                "connector_secret_name": secret_name or "",
            }
        )
        self.save_yaml_file(self.output_config, "output", merge_existing=True)

    def register_model_output(
        self, model_id: str, model_name: str, connector_id: str
    ) -> None:
        """
        Save register model output to a YAML file.

        Args:
            model_id: The registered model ID.
            model_name: The registered model name.
            connector_id: The connector ID associated with the registered model.
        """
        # Update the register_model section
        self.output_config["register_model"].append(
            {
                "model_id": model_id,
                "model_name": model_name,
                "connector_id": connector_id,
            }
        )
        self.save_yaml_file(self.output_config, "output", merge_existing=True)

    def predict_model_output(self, model_id: str, response: str) -> None:
        """
        Save predict model output to a YAML file.

        Args:
            model_id: The model ID used in the prediction.
            response: The response from the prediction.
        """
        # Update the predict_model section
        self.output_config["predict_model"].append(
            {
                "model_id": model_id,
                "response": response,
            }
        )
        self.save_yaml_file(self.output_config, "output", merge_existing=True)

    def get_opensearch_domain_name(
        self, opensearch_domain_endpoint: str
    ) -> Optional[str]:
        """
        Extract the domain name from the OpenSearch endpoint URL.

        Args:
            opensearch_domain_endpoint: URL like 'https://search-domain-abc123.region.es.amazonaws.com'

        Returns:
            Optional[str]:
                - str: Domain name without prefixes and unique suffixes (e.g., 'domain')
                - None: If URL is invalid or empty
        """
        if not opensearch_domain_endpoint:
            return None

        try:
            parsed_url = urlparse(opensearch_domain_endpoint)
            if not parsed_url.hostname:
                return None

            hostname = parsed_url.hostname.split(".")[0]

            # Remove prefix if present
            if hostname.startswith("search-"):
                hostname = hostname[len("search-") :]
            elif hostname.startswith("vpc-"):
                hostname = hostname[len("vpc-") :]

            # Remove unique ID suffix if present
            if "-" in hostname:
                parts = hostname.split("-")
                return "-".join(parts[:-1])
            return hostname

        except Exception as e:
            logger.error(
                f"Error parsing OpenSearch domain endpoint: {opensearch_domain_endpoint}. Error: {str(e)}"
            )
            return None

    def _check_config(
        self,
        config: Dict[str, Any],
        service_type: str,
        opensearch_config: Dict[str, Any],
    ) -> Union[AIConnectorHelper, bool]:
        """
        Check if the configuration is valid for the given service type and OpenSearch configuration.
        """
        aws_credentials = self.config.get("aws_credentials", {})
        opensearch_domain_endpoint = opensearch_config.get("opensearch_domain_endpoint")
        self.opensearch_domain_name = None

        if not opensearch_domain_endpoint:
            logger.warning(
                f"\n{Fore.RED}OpenSearch endpoint not set. Please run setup first.{Style.RESET_ALL}\n"
            )
            return False

        if service_type == self.AMAZON_OPENSEARCH_SERVICE:
            # For managed service, check AWS-specific configurations
            self.opensearch_domain_name = self.get_opensearch_domain_name(
                opensearch_domain_endpoint
            )
            if (
                not opensearch_config.get("opensearch_domain_region")
                or not self.opensearch_domain_name
            ):
                logger.warning(
                    f"{Fore.RED}AWS region or domain name not set. Please run setup first.{Style.RESET_ALL}\n"
                )
                return False

        # Create OpenSearch config
        opensearch_config = OpenSearchDomainConfig(
            opensearch_domain_region=opensearch_config.get("opensearch_domain_region"),
            opensearch_domain_name=self.opensearch_domain_name,
            opensearch_domain_username=opensearch_config.get(
                "opensearch_domain_username"
            ),
            opensearch_domain_password=opensearch_config.get(
                "opensearch_domain_password"
            ),
            opensearch_domain_endpoint=opensearch_config.get(
                "opensearch_domain_endpoint"
            ),
        )

        # Create AWS config
        aws_config = AWSConfig(
            aws_user_name=aws_credentials.get("aws_user_name"),
            aws_role_name=aws_credentials.get("aws_role_name"),
            aws_access_key=aws_credentials.get("aws_access_key"),
            aws_secret_access_key=aws_credentials.get("aws_secret_access_key"),
            aws_session_token=aws_credentials.get("aws_session_token"),
        )

        # Create AIConnectorHelper
        ai_helper = AIConnectorHelper(
            service_type=config.get("service_type"),
            opensearch_config=opensearch_config,
            aws_config=aws_config,
            ssl_check_enabled=config.get("ssl_check_enabled"),
        )
        return ai_helper

    def load_and_check_config(
        self, config_path: str
    ) -> Union[Tuple[AIConnectorHelper, Dict[str, Any], str, Dict[str, Any]], bool]:
        """
        Load and check configuration.

        Args:
            config_path: The file path where the configuration should be loaded from and checked.

        Returns:
            Union[Tuple[AIConnectorHelper, Dict[str, Any], str, Dict[str, Any]], bool]:
                If successful, returns a tuple containing:
                - AIConnectorHelper: Initialized helper instance for AI connectors.
                - Dict[str, Any]: Complete configuration dictionary.
                - str: Service type extracted from configuration.
                - Dict[str, Any]: OpenSearch domain configuration.

                If any validation fails, returns False.

        """
        # Load configuration
        config = self.load_config(config_path)
        if not config:
            return False

        service_type = config.get("service_type")
        opensearch_config = self.config.get("opensearch_config", {})

        # Check configuration validity
        ai_helper = self._check_config(config, service_type, opensearch_config)
        if not ai_helper:
            return False
        return ai_helper, config, service_type, opensearch_config



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

import logging
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from colorama import Fore, Style, init
from rich.console import Console

from opensearch_py_ml.ml_commons.cli.ai_connector_helper import AIConnectorHelper
from opensearch_py_ml.ml_commons.cli.cli_base import CLIBase
from opensearch_py_ml.ml_commons.cli.connector_list import (
    ConnectorInfo,
    ConnectorList,
    ModelInfo,
)

# Initialize Rich console for enhanced CLI outputs
console = Console()

# Initialize colorama for colored terminal output
init(autoreset=True)

# Configure the logger for this module
logger = logging.getLogger(__name__)


class ConnectorManager(CLIBase):
    """
    Handles the connector operation.
    """

    def __init__(self):
        """
        Initialize the ConnectorManager class.
        """
        super().__init__()
        self.config = {}
        self.connector_list = ConnectorList()

    def get_connectors(self, service_type: str) -> List[ConnectorInfo]:
        """
        Get connectors for specific service type.

        Args:
            service_type: The type of service to get connectors for. Either self.OPEN_SOURCE or self.AMAZON_OPENSEARCH_SERVICE.

        Returns:
            List[Connector]: List of connector objects for the specified service type.

        Raises:
            ValueError: If the service_type is not recognized.
        """
        if service_type == self.OPEN_SOURCE:
            return self.connector_list._opensource_connectors
        elif service_type == self.AMAZON_OPENSEARCH_SERVICE:
            return self.connector_list._managed_connectors
        else:
            raise ValueError(f"Unknown service type: {service_type}")

    def print_available_connectors(self, service_type: str) -> None:
        """
        Print available connectors for specific service type.

        Args:
            service_type: The type of service to list connectors for.
        """
        connectors = self.get_connectors(service_type)
        if not connectors:
            logger.warning(f"\nNo connectors available for {service_type}")
            return

        service_name = "Amazon OpenSearch Service"
        if service_type == self.OPEN_SOURCE:
            service_name = "OpenSearch"
        print(f"\nPlease select a supported connector to create in {service_name}:")
        for connector in connectors:
            print(f"{connector.id}. {connector.name}")
        max_choice = len(connectors)
        print(f"Enter your choice (1-{max_choice}): ", end="")

    def get_connector_by_id(
        self, connector_id: int, service_type: str
    ) -> ConnectorInfo:
        """
        Get connector by ID for specific service type.

        Args:
            connector_id: The numeric identifier of the connector.
            service_type: The type of service the connector belongs to.

        Returns:
            ConnectorInfo: The connector object matching the specified ID.

        Raises:
            ValueError: If no connector is found with the specified ID.
        """
        connectors = self.get_connectors(service_type)
        for connector in connectors:
            if connector.id == connector_id:
                return connector
        raise ValueError

    def get_connector_by_name(self, name: str, service_type: str) -> ConnectorInfo:
        """
        Get connector by name for specific service type.

        Args:
            name: The name of the connector.
            service_type: The type of service the connector belongs to.

        Returns:
            ConnectorInfo: he connector object matching the specified name.

        Raises:
            ValueError: If no connector is found with the specified name.
        """
        connectors = self.get_connectors(service_type)
        for connector in connectors:
            if connector.name.lower() == name.lower():
                return connector
        raise ValueError

    def get_available_models(
        self, connector_name: str, service_type: str
    ) -> List[ModelInfo]:
        """
        Get available models for a specific connector.

        Args:
            connector-name: Name of the connector to get models for.
            service_type: The type of service the connector belongs to.

        Returns:
            List[ModelInfo]: List of ModelInfo objects containing available models.
                Each ModelInfo object has:
                - id (str): Unique identifier for the model.
                - name (str): Display name of the model.
                Returns empty list if connector is not found.
        """
        connectors = (
            self.connector_list._managed_connectors
            if service_type == self.AMAZON_OPENSEARCH_SERVICE
            else self.connector_list._opensource_connectors
        )
        for connector in connectors:
            if connector.name == connector_name:
                return connector.available_models
        return []

    def get_connector_class(self, connector_name: str) -> Optional[Type[Any]]:
        """
        Import and return the connector class.

        Args:
            connector_name: Name of the connector to get the class for.

        Returns:
            Optional[Type[Any]]: The connector class if found and successfully imported,
            None if:
                - Connector information not found
                - Module import fails
                - Class not found in module
        """
        connector = self.get_connector_info(connector_name)
        if connector:
            module_path = (
                f"opensearch_py_ml.ml_commons.cli.ml_models.{connector.file_name}"
            )
            module = __import__(module_path, fromlist=[connector.connector_class])
            return getattr(module, connector.connector_class)
        return None

    def get_connector_info(self, connector_name: str) -> Optional[ConnectorInfo]:
        """
        Get connector information by name.

        Args:
            connector_name: Name of the connector to find information for.

        Returns:
            Optional[ConnectorInfo]: ConnectorInfo object if found.
            None if no matching connector is found.
        """
        for connector in (
            self.connector_list._managed_connectors
            + self.connector_list._opensource_connectors
        ):
            if connector.name == connector_name:
                return connector
        return None

    def _get_connector_config_params(
        self, connector_config: Dict[str, Any]
    ) -> Dict[str, Optional[Any]]:
        """
        Get connector configuration parameters.
        """
        param_keys = [
            "access_token",
            "api_key",
            "aws_access_key",
            "aws_secret_access_key",
            "aws_session_token",
            "connector_body",
            "connector_name",
            "connector_role_inline_policy",
            "connector_role_prefix",
            "connector_secret_name",
            "endpoint_arn",
            "endpoint_url",
            "model_id",
            "model_name",
            "project_id",
            "region",
            "required_policy",
            "required_secret",
        ]

        params = {}
        for key in param_keys:
            params[key] = connector_config.get(key)
        return params

    def create_model_instance(
        self,
        connector_info: ConnectorInfo,
        connector_class: Type[Any],
        opensearch_config: Dict[str, Any],
        config: Dict[str, Any],
    ) -> Any:
        """
        Create model instance.

        Args:
            connector_info: Connector information object.
            connector_class: The class to instantiate for the model.
            opensearch_config: OpenSearch domain configuration.
            config: General configuration parameters.

        Returns:
            Any: An instance of the connector_class initialized with the required parameters.
        """
        init_kwargs = {}
        for param in connector_info.init_params:
            # Check opensearch_config
            if param in opensearch_config and opensearch_config.get(param) is not None:
                init_kwargs[param] = opensearch_config.get(param)
            # Check config
            elif param in config and config.get(param) is not None:
                init_kwargs[param] = config.get(param)

        return connector_class(**init_kwargs)

    def create_connector_instance(
        self,
        connector_config_path: Optional[str],
        connector_config_params: Optional[Dict[str, Any]],
        connector_info: ConnectorInfo,
        opensearch_config: Dict[str, Any],
        config: Dict[str, Any],
        model: Any,
        ai_helper: AIConnectorHelper,
    ) -> Any:
        """
        Create connector instance.

        Args:
            connector_config_path: Path to connector configuration file.
            connector_config_params: Parameters from connector configuration.
            connector_info: Connector information object.
            opensearch_config: OpenSearch domain configuration.
            config: General configuration parameters.
            model: Model instance to use for connector creation.
            ai_helper: Helper instance for AI connector operations.

        Returns:
            Any: Created connector instance.
        """
        connector_kwargs = {}
        for param in connector_info.connector_params:
            # Check connector_params from connector config
            if (
                connector_config_path
                and param in connector_config_params
                and connector_config_params[param] is not None
            ):
                connector_kwargs[param] = connector_config_params[param]
            # Check opensearch_config
            elif (
                param in opensearch_config and opensearch_config.get(param) is not None
            ):
                connector_kwargs[param] = opensearch_config.get(param)
            # Check config
            elif param in config and config.get(param) is not None:
                connector_kwargs[param] = config.get(param)

        return model.create_connector(
            ai_helper, self.connector_output, **connector_kwargs
        )

    def initialize_create_connector(
        self, connector_config_path: Optional[str] = None
    ) -> Union[Tuple[bool, str], bool]:
        """
        Orchestrates the entire connector creation process.

        Args:
            connector_config_path: Path to connector configuration file. If None, will prompt for setup configuration path.

        Returns:
            Union[Tuple[bool, str], bool]:
                - If successful: Tuple(True, setup_config_path)
                - If failed: False
        """
        try:
            # Check if connector config file path is given in the command
            if connector_config_path:
                connector_config = self.load_config(connector_config_path, "connector")
                setup_config_path = connector_config.get("setup_config_path")
                if not connector_config:
                    logger.error(
                        f"{Fore.RED}No connector configuration found.{Style.RESET_ALL}\n"
                    )
                    return False
            else:
                setup_config_path = input(
                    "\nEnter the path to your existing setup configuration file: "
                ).strip()

            # Load and check configuration
            config_result = self.load_and_check_config(setup_config_path)
            if not config_result:
                return False
            ai_helper, config, service_type, opensearch_config = config_result

            # Set the initial value of the connector creation result to False
            result = False

            # Retrieve the connector information if the connector config path is given in the command
            if connector_config_path:
                params = self._get_connector_config_params(connector_config)
                connector_name = params.pop("connector_name")

                try:
                    connector_info = self.get_connector_by_name(
                        connector_name, service_type
                    )
                except ValueError:
                    logger.error(
                        f"{Fore.YELLOW}Invalid connector choice. Operation cancelled.{Style.RESET_ALL}"
                    )
                    return False
            else:
                self.print_available_connectors(service_type)
                try:
                    choice = int(input().strip())
                    connector_info = self.get_connector_by_id(choice, service_type)
                except ValueError:
                    logger.error(
                        f"{Fore.YELLOW}Invalid connector choice. Operation cancelled.{Style.RESET_ALL}"
                    )
                    return False

            connector_name = connector_info.name
            connector_class = self.get_connector_class(connector_name)

            # Create model instance
            self.model = self.create_model_instance(
                connector_info, connector_class, opensearch_config, config
            )

            # Create connector instance
            result = self.create_connector_instance(
                connector_config_path,
                params if connector_config_path else None,
                connector_info,
                opensearch_config,
                config,
                self.model,
                ai_helper,
            )

            return result, setup_config_path
        except Exception as e:
            logger.error(
                f"{Fore.RED}Error creating connector: {str(e)}{Style.RESET_ALL}"
            )
            return False



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

from dataclasses import dataclass
from typing import List


@dataclass
class ModelInfo:
    """
    Information about a specific model within a connector
    """

    id: str
    name: str


@dataclass
class ConnectorInfo:
    """
    Information about a connector and its available models
    """

    id: int
    name: str
    file_name: str
    connector_class: str
    init_params: List[str]
    connector_params: List[str]
    available_models: List[ModelInfo] = None


class ConnectorList:
    def __init__(self):
        # TODO: Add more supported connectors for both AOS and open-source
        # List of supported connectors in open-source service
        self._opensource_connectors: List[ConnectorInfo] = [
            ConnectorInfo(
                id=1,
                name="Aleph Alpha",
                file_name="aleph_alpha_model",
                connector_class="AlephAlphaModel",
                init_params=[],
                connector_params=["model_name", "api_key", "connector_body"],
                available_models=[
                    ModelInfo(id="1", name="Luminous-Base embedding model"),
                    ModelInfo(id="2", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=2,
                name="Amazon Bedrock",
                file_name="bedrock_model",
                connector_class="BedrockModel",
                init_params=["opensearch_domain_region", "service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "region",
                    "model_name",
                    "connector_body",
                    "aws_access_key",
                    "aws_secret_access_key",
                    "aws_session_token",
                ],
                available_models=[
                    ModelInfo(id="1", name="AI21 Labs Jurassic-2 Mid"),
                    ModelInfo(id="2", name="Anthropic Claude v2"),
                    ModelInfo(id="3", name="Anthropic Claude v3"),
                    ModelInfo(id="4", name="Anthropic Claude v3.7"),
                    ModelInfo(id="5", name="Cohere Embed Model v3 - English"),
                    ModelInfo(id="6", name="Cohere Embed Model v3 - Multilingual"),
                    ModelInfo(id="7", name="Titan Text Embedding"),
                    ModelInfo(id="8", name="Titan Multimodal Embedding"),
                    ModelInfo(id="9", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=3,
                name="Amazon Bedrock Converse",
                file_name="bedrock_converse_model",
                connector_class="BedrockConverseModel",
                init_params=["opensearch_domain_region", "service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "region",
                    "model_name",
                    "connector_body",
                    "aws_access_key",
                    "aws_secret_access_key",
                    "aws_session_token",
                ],
                available_models=[
                    ModelInfo(id="1", name="Anthropic Claude 3 Sonnet"),
                    ModelInfo(id="2", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=4,
                name="Amazon Comprehend",
                file_name="comprehend_model",
                connector_class="ComprehendModel",
                init_params=["opensearch_domain_region", "service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "region",
                    "model_name",
                    "connector_body",
                    "aws_access_key",
                    "aws_secret_access_key",
                    "aws_session_token",
                ],
                available_models=[
                    ModelInfo(id="1", name="Metadata embedding model"),
                    ModelInfo(id="2", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=5,
                name="Amazon Textract",
                file_name="textract_model",
                connector_class="TextractModel",
                init_params=["opensearch_domain_region", "service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "region",
                    "model_name",
                    "connector_body",
                    "aws_access_key",
                    "aws_secret_access_key",
                    "aws_session_token",
                ],
                available_models=[
                    ModelInfo(id="1", name="Amazon Textract model"),
                    ModelInfo(id="2", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=6,
                name="Azure OpenAI",
                file_name="azure_openai_model",
                connector_class="AzureOpenAIModel",
                init_params=[],
                connector_params=[
                    "model_name",
                    "api_key",
                    "resource_name",
                    "deployment_name",
                    "api_version",
                    "connector_body",
                ],
                available_models=[
                    ModelInfo(id="1", name="Chat completion model"),
                    ModelInfo(id="2", name="Embedding model"),
                    ModelInfo(id="3", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=7,
                name="Amazon SageMaker",
                file_name="sagemaker_model",
                connector_class="SageMakerModel",
                init_params=["opensearch_domain_region", "service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "region",
                    "model_name",
                    "endpoint_arn",
                    "endpoint_url",
                    "connector_body",
                    "aws_access_key",
                    "aws_secret_access_key",
                    "aws_session_token",
                ],
                available_models=[
                    ModelInfo(id="1", name="Embedding model"),
                    ModelInfo(id="2", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=8,
                name="Cohere",
                file_name="cohere_model",
                connector_class="CohereModel",
                init_params=["service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "model_name",
                    "api_key",
                    "connector_body",
                    "connector_secret_name",
                ],
                available_models=[
                    ModelInfo(id="1", name="Chat model"),
                    ModelInfo(id="2", name="Embedding model"),
                    ModelInfo(id="3", name="Multi-modal embedding model"),
                    ModelInfo(id="4", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=9,
                name="DeepSeek",
                file_name="deepseek_model",
                connector_class="DeepSeekModel",
                init_params=["service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "model_name",
                    "api_key",
                    "connector_body",
                    "connector_secret_name",
                ],
                available_models=[
                    ModelInfo(id="1", name="DeepSeek Chat model"),
                    ModelInfo(id="2", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=10,
                name="Google Cloud Platform",
                file_name="gcp_model",
                connector_class="GCPModel",
                init_params=[],
                connector_params=[
                    "model_name",
                    "project_id",
                    "model_id",
                    "access_token",
                    "connector_body",
                ],
                available_models=[
                    ModelInfo(id="1", name="VertexAI embedding model"),
                    ModelInfo(id="2", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=11,
                name="OpenAI",
                file_name="openai_model",
                connector_class="OpenAIModel",
                init_params=["service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "model_name",
                    "api_key",
                    "connector_body",
                    "connector_secret_name",
                ],
                available_models=[
                    ModelInfo(id="1", name="Chat model"),
                    ModelInfo(id="2", name="Completion model"),
                    ModelInfo(id="3", name="Embedding model"),
                    ModelInfo(id="4", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=12,
                name="Custom connector",
                file_name="custom_connector",
                connector_class="CustomConnector",
                init_params=["service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "connector_secret_name",
                    "connector_role_inline_policy",
                    "api_key",
                    "required_policy",
                    "required_secret",
                    "connector_body",
                ],
                available_models=[],
            ),
        ]
        # List of supported connectors in managed service (AOS)
        self._managed_connectors: List[ConnectorInfo] = [
            ConnectorInfo(
                id=1,
                name="Amazon Bedrock",
                file_name="bedrock_model",
                connector_class="BedrockModel",
                init_params=["opensearch_domain_region", "service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "region",
                    "model_name",
                    "connector_body",
                    "aws_access_key",
                    "aws_secret_access_key",
                    "aws_session_token",
                ],
                available_models=[
                    ModelInfo(id="1", name="AI21 Labs Jurassic-2 Mid"),
                    ModelInfo(id="2", name="Anthropic Claude v2"),
                    ModelInfo(id="3", name="Anthropic Claude v3"),
                    ModelInfo(id="4", name="Anthropic Claude v3.7"),
                    ModelInfo(id="5", name="Cohere Embed Model v3 - English"),
                    ModelInfo(id="6", name="Cohere Embed Model v3 - Multilingual"),
                    ModelInfo(id="7", name="Titan Text Embedding"),
                    ModelInfo(id="8", name="Titan Multimodal Embedding"),
                    ModelInfo(id="9", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=2,
                name="Amazon Bedrock Converse",
                file_name="bedrock_converse_model",
                connector_class="BedrockConverseModel",
                init_params=["opensearch_domain_region", "service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "region",
                    "model_name",
                    "connector_body",
                    "aws_access_key",
                    "aws_secret_access_key",
                    "aws_session_token",
                ],
                available_models=[
                    ModelInfo(id="1", name="Anthropic Claude 3 Sonnet"),
                    ModelInfo(id="2", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=3,
                name="Amazon Comprehend",
                file_name="comprehend_model",
                connector_class="ComprehendModel",
                init_params=["opensearch_domain_region", "service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "region",
                    "model_name",
                    "connector_body",
                    "aws_access_key",
                    "aws_secret_access_key",
                    "aws_session_token",
                ],
                available_models=[
                    ModelInfo(id="1", name="Metadata embedding model"),
                    ModelInfo(id="2", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=4,
                name="Amazon SageMaker",
                file_name="sagemaker_model",
                connector_class="SageMakerModel",
                init_params=["opensearch_domain_region", "service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "region",
                    "model_name",
                    "endpoint_arn",
                    "endpoint_url",
                    "connector_body",
                    "aws_access_key",
                    "aws_secret_access_key",
                    "aws_session_token",
                ],
                available_models=[
                    ModelInfo(id="1", name="DeepSeek R1 model"),
                    ModelInfo(id="2", name="Embedding model"),
                    ModelInfo(id="3", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=5,
                name="Amazon Textract",
                file_name="textract_model",
                connector_class="TextractModel",
                init_params=["opensearch_domain_region", "service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "region",
                    "model_name",
                    "connector_body",
                    "aws_access_key",
                    "aws_secret_access_key",
                    "aws_session_token",
                ],
                available_models=[
                    ModelInfo(id="1", name="Amazon Textract model"),
                    ModelInfo(id="2", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=6,
                name="Cohere",
                file_name="cohere_model",
                connector_class="CohereModel",
                init_params=["service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "model_name",
                    "api_key",
                    "connector_body",
                    "connector_secret_name",
                ],
                available_models=[
                    ModelInfo(id="1", name="Embedding model"),
                    ModelInfo(id="2", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=7,
                name="DeepSeek",
                file_name="deepseek_model",
                connector_class="DeepSeekModel",
                init_params=["service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "model_name",
                    "api_key",
                    "connector_body",
                    "connector_secret_name",
                ],
                available_models=[
                    ModelInfo(id="1", name="DeepSeek Chat model"),
                    ModelInfo(id="2", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=8,
                name="OpenAI",
                file_name="openai_model",
                connector_class="OpenAIModel",
                init_params=["service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "model_name",
                    "api_key",
                    "connector_body",
                    "connector_secret_name",
                ],
                available_models=[
                    ModelInfo(id="1", name="Embedding model"),
                    ModelInfo(id="2", name="Custom model"),
                ],
            ),
            ConnectorInfo(
                id=9,
                name="Custom connector",
                file_name="custom_connector",
                connector_class="CustomConnector",
                init_params=["service_type"],
                connector_params=[
                    "connector_role_prefix",
                    "connector_secret_name",
                    "connector_role_inline_policy",
                    "api_key",
                    "required_policy",
                    "required_secret",
                    "connector_body",
                ],
                available_models=[],
            ),
        ]



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

"""wavelet_piecewise_const

Tools for fast piecewise-constant signal approximation from Haar wavelet reconstruction.

Functions
---------
piecewise_const_to_segs
    Segment-based representation of a piecewise constant signal.

blockwise_dp_segment
    DP-based piecewise constant approximation of a piecewise-constant signal.

blockwise_seg_to_approx
    Convert output of blockwise_dp_segment to a signal-length approximation.

lavielle_criterion
    Criterion for selecting number of segments.

wavelet_piecewise_const
    Fast piecewise constant approximation via Haar wavelet reconstruction.

tranform_piecewise_const
    Convert piecewise constant signal to activity score.
"""
import math
from typing import Dict, List, Tuple

import torch
import torch.nn.functional as F

from .wavelet_tools import haar_approx


def wavelet_piecewise_const(
    data: torch.Tensor, m: int, max_dp_levels: int, is_haar_approx: bool
) -> torch.Tensor:
    """
    Fast, wavelet-based piecewise constant signal approximation for row m of the M x T input data tensor.
    :param data:
    :type data: torch.Tensor
    :param m:
    :type m: int
    :param max_dp_levels
    :type max_dp_levels: int
    :param is_haar_approx
    :type is_haar_approx: bool
    :return:
    :rtype: torch.Tensor
    """
    seq = data[m, :]
    if (torch.abs(torch.diff(seq)) > 0).sum() == 0:
        return seq  # avoid cost and numerical errors from needless approx
    elif is_haar_approx:  # compute Haar piecewise constant approx
        N = len(seq)
        J = math.ceil(math.log(N) / math.log(2))
        wav_approx = haar_approx(seq, truncate=J * max_dp_levels)
    else:
        wav_approx = seq  # no wavelet approximation -> slower DP

    # refine with block wise DP
    segs, costs = block_wise_dp_segment(wav_approx, max_dp_levels)

    # select segmentation level with lavielle criterion
    nseg, _ = lavielle_criterion(costs[:, -1])

    return block_wise_seg_to_approx(wav_approx, segs[nseg])


def block_wise_dp_segment(
    data: torch.Tensor, max_lvls: int
) -> Tuple[Dict[int, List[int]], torch.Tensor]:
    """
    Piecewise constant approximation to a piecewise constant signal, i.e. the best approximation at a more granular
        (fewer segments) resolution, via dynamic programming.
    :param data:
    :type data: torch.Tensor
    :param max_lvls:
    :type max_lvls: int
    :return:
    :rtype: Tuple[Dict[int, List[int]], torch.Tensor]
    """

    vals, counts = piecewise_const_to_segs(data)
    ssum: List[float] = [
        float(x) for x in torch.cumsum(vals * counts, 0)
    ]  # need to do this as tensor.tolist() gives a torch.jit runtime error
    sqsum: List[float] = [
        float(x) for x in torch.cumsum(torch.pow(vals, 2) * counts, 0)
    ]
    scounts: List[float] = [float(x) for x in torch.cumsum(counts, 0)]

    N = len(vals)
    segs = {int(i + 1): [0] for i in range(max_lvls)}
    costs = torch.zeros((max_lvls, N))

    # initialize cost and segmentation in base case w/ b=0 (no jumps)
    segs[1] = [0, int(N)]
    best_last_segs = [[0]] * N

    costs[0, :] = torch.zeros(N)
    for i in range(N):
        costs[0, i] = sqerr_seg_list(0, i, ssum, sqsum, scounts)

    # recursion
    for b in range(1, max_lvls):
        best_segs: List[List[int]] = []

        for i in range(N):
            if i <= b:  # more breaks than segs in data -> 0 cost
                seg_i: List[int] = []
                for x in range(
                    i + 1
                ):  # expand list comp here as torchscript can't handle it
                    seg_i += [x]
                best_segs += [seg_i]
                costs[b, i] = 0

            else:  # DP recursion
                addcosts = torch.zeros(i)
                for j in range(i):
                    addcosts[j] = sqerr_seg_list(j + 1, i, ssum, sqsum, scounts)
                new_jump_costs = costs[b - 1, :i] + addcosts

                new_jump_ix = int(torch.argmin(new_jump_costs))
                new_jump_cost = new_jump_costs[new_jump_ix]

                best_segs += [best_last_segs[new_jump_ix] + [new_jump_ix + 1]]
                costs[b, i] = new_jump_cost

        best_last_segs = best_segs
        segs[b + 1] = best_segs[-1] + [int(N)]

    return segs, costs


def lavielle_criterion(costs: torch.Tensor) -> Tuple[int, torch.Tensor]:
    """
    Criterion to automatically select number of segments in a piecewise constant approximation.
        Reference: https://hal.inria.fr/inria-00070662/document
    :param costs:
    :type costs: torch.Tensor
    :return:
    :rtype: Tuple[int, torch.Tensor]
    """
    s = 0.75  # hardcode cutoff at paper suggestion
    kmax = len(costs)
    normcosts = (costs[-1] - costs) / (costs[-1] - costs[0]) * (kmax - 1) + 1

    # catch NaNs arising from div-by-zero above
    # only happens when cost[1] = cost[kmax], so correct choice is 1 segment
    normcosts[normcosts.isnan()] = -1

    d = torch.zeros(len(costs))
    d[0] = (
        4 * torch.max(torch.abs(normcosts)) + 1.1 * s
    )  # upper bound and always strictly > s
    for i in range(1, len(costs) - 1):
        d[i] = (
            normcosts[i - 1] - 2 * normcosts[i] + normcosts[i + 1]
        )  # discrete 2nd deriv

    return int(torch.max(torch.nonzero(d > s)[:, 0]) + 1), d  # largest k s.t. d[k]>s


def block_wise_seg_to_approx(data: torch.Tensor, seg: List[int]) -> torch.Tensor:
    """
    Convert blockwise DP segmentation output to piecewise constant approximation.
    :param costs:
    :type costs: torch.Tensor
    :return:
    :rtype: Tuple[int, torch.Tensor]
    """
    vals, counts = piecewise_const_to_segs(data)
    scounts = F.pad(
        torch.cumsum(counts, 0), (1, 0)
    )  # prepend zero to get indices right

    apx = torch.zeros(len(data))
    for i in range(1, len(seg)):
        e = seg[i]  # DP segments are left-closed, right-open
        s = seg[i - 1]

        eix = scounts[e]
        six = scounts[s]

        apx[six:eix] = torch.mean(data[six:eix])

    return apx


def piecewise_const_to_segs(signal: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Represent a piecewise-constant signal in terms of its segments, each given by a value and length.
    :param signal:
    :type signal: torch.Tensor
    :return: output -> the output list of unique scalar elements, counts ->
        represents the number of occurrences for each unique value or tensor.
    :rtype: torch.Tensor, torch.Tensor
    """
    vals, counts = torch.unique_consecutive(signal, return_counts=True)
    return vals, counts


def sqerr_seg_list(
    st: int, e: int, ssum: List[float], sqsum: List[float], scounts: List[float]
) -> float:
    """
    Blockwise version of square loss.
    :param st:
    :type st: int
    :param e:
    :type e: int
    :param ssum:
    :type ssum: List[float]
    :param sqsum:
    :type sqsum: List[float]
    :param scounts:
    :type scounts: List[float]
    :return:
    :rtype:
    """

    assert e >= st
    s = ssum[e] - ssum[st - 1] if st > 0 else ssum[e]
    sq = sqsum[e] - sqsum[st - 1] if st > 0 else sqsum[e]
    n = scounts[e] - scounts[st - 1] if st > 0 else scounts[e]
    return sq - (1 / n) * s**2


def transform_piecewise_const(sig: torch.Tensor, bandwidth: float) -> torch.Tensor:
    """
    Convert piecewise constant approximation to activity score.
    :param sig: signal
    :type sig: torch.Tensor
    :param bandwidth:
    :type bandwidth: float
    :return:
    :rtype: torch.Tensor
    """
    sigdiff = torch.abs(torch.diff(sig))

    # logic required to avoid bad representation of spikes
    chg = torch.where(sigdiff > 0)[0]
    for ix in chg:
        if torch.any(chg == ix + 1):
            sigdiff[ix + 1] = 0

    m = torch.linalg.norm(sigdiff)
    if m != 0:
        sigdiff /= m  # note this may amplify small variations

    blurred = gconv(sigdiff, bandwidth)
    return blurred


def gconv(data: torch.Tensor, bandwidth: float) -> torch.Tensor:
    """
    1-D convolution with boundary padding.
    :param data:
    :type data: torch.Tensor
    :param bandwidth:
    :type bandwidth: float
    :return:
    :rtype: torch.Tensor
    """

    filt = gfilt(bandwidth)
    padlen = len(filt) // 2
    padded = boundary_pad(data, (padlen, padlen))

    spdat = torch.unsqueeze(torch.unsqueeze(padded, 0), 0)
    sfilt = torch.unsqueeze(torch.unsqueeze(filt, 0), 0)
    res = F.conv1d(spdat, sfilt)
    return res.squeeze()


def gfilt(sigma: float, truncate: float = 4.0) -> torch.Tensor:
    """
    Pytorch Gaussian filter with truncation.
    :param sigma: sigma
    :type sigma: float
    :param truncate:
    :type truncate: float, default value 4.0
    :return:
    :rtype: torch.Tensor
    """

    lw = int(truncate * sigma + 0.5)
    xs = torch.arange(-lw, lw + 1)
    sigma2 = math.pow(sigma, 2)
    weights = torch.exp(-0.5 / sigma2 * torch.pow(xs, 2))
    weights = weights / torch.sum(weights)

    return weights


def boundary_pad(sig: torch.Tensor, lens: List[int]) -> torch.Tensor:
    """
    Pad signal by extending value at boundary ('nearest' in scipy).
    :param sig: signal
    :type sig: float
    :param lens:
    :type lens: List[int]
    :return:
    :rtype: torch.Tensor
    """

    lpadded = F.pad(sig, (lens[0], 0), mode="constant", value=sig[0])
    padded = F.pad(lpadded, (0, lens[1]), mode="constant", value=sig[-1])
    return padded



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.


#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

# flake8: noqa

from codecs import open
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))
about = {}
with open(path.join(here, "opensearch_py_ml", "_version.py"), "r", "utf-8") as f:
    exec(f.read(), about)

CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "License :: OSI Approved :: Apache Software License",
    "Environment :: Console",
    "Operating System :: OS Independent",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: 3.13",
    "Topic :: Scientific/Engineering",
]

# Remove all raw HTML from README for long description
with open(path.join(here, "README.md"), "r", "utf-8") as f:
    lines = f.read().split("\n")
    last_html_index = 0
    for i, line in enumerate(lines):
        if line == "</p>":
            last_html_index = i + 1
    long_description = "\n".join(lines[last_html_index:])

extras = {}
extras["all"] = list({dep for deps in extras.values() for dep in deps})

setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=about["__url__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    maintainer=about["__maintainer__"],
    maintainer_email=about["__maintainer_email__"],
    license="Apache-2.0",
    classifiers=CLASSIFIERS,
    keywords="Opensearch opensearch_py_ml pandas python",
    packages=find_packages(include=["opensearch_py_ml", "opensearch_py_ml.*"]),
    project_urls={
        "Source Code": "https://github.com/opensearch-project/opensearch-py-ml",
        "Issue Tracker": "https://github.com/opensearch-project/opensearch-py-ml/issues",
    },
    install_requires=[
        "opensearch-py>=2",
        "pandas>=1.5.2,<2.3,!=2.1.0",
        "matplotlib>=3.6.0,<4",
        "numpy>=2.3.2",
        "deprecated>=1.2.14,<2",
        "boto3>=1.26.0",
        "botocore>=1.29.0",
        "requests>=2.28.0",
        "requests-aws4auth>=1.1.0",
        "colorama>=0.4.6",
        "rich>=13.5.2",
        "pyyaml>=6.0.2",
    ],
    python_requires=">=3.11",
    package_data={
        "opensearch_py_ml": [
            "py.typed",
            "ml_commons/cli/ml_models/*.py",
            "ml_commons/cli/ml_models/__init__.py",
        ]
    },
    include_package_data=True,
    zip_safe=False,
    extras_require=extras,
    # Entry points for console scripts
    entry_points={
        "console_scripts": [
            "opensearch-ml=opensearch_py_ml.ml_commons.cli.ml_cli:main",
        ],
    },
)



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.


class OpenSearchDomainConfig:
    """
    Base class for OpenSearch domain configuration.
    """

    def __init__(
        self,
        opensearch_domain_region: str,
        opensearch_domain_name: str,
        opensearch_domain_username: str,
        opensearch_domain_password: str,
        opensearch_domain_endpoint: str,
    ):
        """
        Initialize OpenSearch domain configuration.
        """
        self.opensearch_domain_region = opensearch_domain_region
        self.opensearch_domain_name = opensearch_domain_name
        self.opensearch_domain_username = opensearch_domain_username
        self.opensearch_domain_password = opensearch_domain_password
        self.opensearch_domain_endpoint = opensearch_domain_endpoint



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

import json
from typing import Any, Callable, Dict, Optional

from colorama import Fore, Style

from opensearch_py_ml.ml_commons.cli.ai_connector_helper import AIConnectorHelper
from opensearch_py_ml.ml_commons.cli.ml_models.model_base import ModelBase


class CohereModel(ModelBase):
    def __init__(
        self,
        service_type,
    ):
        """
        Initializes the Cohere model with necessary configuration.
        """
        self.service_type = service_type

    def _get_connector_body(self, model_type: str) -> Dict[str, Any]:
        """
        Get the connectory body.
        """
        connector_configs = {
            self.AMAZON_OPENSEARCH_SERVICE: {
                "1": {
                    "name": "Cohere Embedding Model Connector",
                    "description": "Connector for Cohere embedding model",
                    "model": "embed-english-v3.0",
                    "url": "https://api.cohere.ai/v1/embed",
                    "request_body": '{ "texts": ${parameters.texts}, "truncate": "${parameters.truncate}", "model": "${parameters.model}", "input_type": "${parameters.input_type}" }',
                    "pre_process_function": "connector.pre_process.cohere.embedding",
                    "post_process_function": "connector.post_process.cohere.embedding",
                    "parameters": {
                        "input_type": "search_document",
                        "truncate": "END",
                    },
                },
                "2": "Custom model",
            },
            self.OPEN_SOURCE: {
                "1": {
                    "name": "Cohere Chat model",
                    "description": "The connector to Cohere's public chat API",
                    "model": "command",
                    "credential": {"cohere_key": "${credential}"},
                    "url": "https://api.cohere.ai/v1/chat",
                    "request_body": '{ "message": "${parameters.message}", "model": "${parameters.model}" }',
                    "parameters": {},
                },
                "2": {
                    "name": "Cohere Embedding model",
                    "description": "The connector to Cohere's public embed API",
                    "model": "embed-english-v3.0",
                    "credential": {"cohere_key": "${credential}"},
                    "url": "https://api.cohere.ai/v1/embed",
                    "request_body": '{ "texts": ${parameters.texts}, "truncate": "${parameters.truncate}", "model": "${parameters.model}", "input_type": "${parameters.input_type}" }',
                    "pre_process_function": "connector.pre_process.cohere.embedding",
                    "post_process_function": "connector.post_process.cohere.embedding",
                    "parameters": {
                        "input_type": "search_document",
                        "truncate": "END",
                    },
                },
                "3": {
                    "name": "Cohere Image Embed model",
                    "description": "The connector to Cohere's public embed API",
                    "model": "embed-english-v3.0",
                    "credential": {"cohere_key": "${credential}"},
                    "url": "https://api.cohere.ai/v1/embed",
                    "request_body": '{ "images": ${parameters.images}, "truncate": "${parameters.truncate}", "model": "${parameters.model}", "input_type": "${parameters.input_type}" }',
                    "pre_process_function": "connector.pre_process.cohere.multimodal_embedding",
                    "post_process_function": "connector.post_process.cohere.embedding",
                    "parameters": {
                        "input_type": "image",
                        "truncate": "END",
                    },
                },
                "4": "Custom model",
            },
        }

        service_configs = connector_configs.get(self.service_type)

        # Handle custom model or invalid choice
        if (
            model_type not in service_configs
            or service_configs[model_type] == "Custom model"
        ):
            if model_type not in service_configs:
                print(
                    f"\n{Fore.YELLOW}Invalid choice. Defaulting to 'Custom model'.{Style.RESET_ALL}"
                )
            return self.input_custom_model_details(external=True)

        config = service_configs[model_type]

        # Base parameters that all connectors need
        base_parameters = {"model": config["model"]}

        # Merge with model-specific parameters if any
        parameters = {**base_parameters, **config.get("parameters", {})}

        # Return the connector body
        return {
            "name": config["name"],
            "description": config["description"],
            "version": 1,
            "protocol": "http",
            "parameters": parameters,
            **({"credential": config["credential"]} if "credential" in config else {}),
            "actions": [
                {
                    "action_type": "predict",
                    "method": "POST",
                    "headers": {
                        "Authorization": "${auth}",
                        "Request-Source": "unspecified:opensearch",
                    },
                    "url": config["url"],
                    "request_body": config["request_body"],
                    **(
                        {"pre_process_function": config["pre_process_function"]}
                        if "pre_process_function" in config
                        else {}
                    ),
                    **(
                        {"post_process_function": config["post_process_function"]}
                        if "post_process_function" in config
                        else {}
                    ),
                }
            ],
        }

    def create_connector(
        self,
        helper: AIConnectorHelper,
        save_config_method: Callable[[str, Dict[str, Any]], None],
        connector_role_prefix: Optional[str] = None,
        model_name: Optional[str] = None,
        api_key: Optional[str] = None,
        connector_body: Optional[Dict[str, Any]] = None,
        connector_secret_name: Optional[str] = None,
    ) -> bool:
        """
        Create Cohere connector.

        Args:
            helper: Helper instance for OpenSearch connector operations.
            save_config_method: Method to save connector configuration after creation.
            connector_role_prefix (optional): Prefix for role names.
            model_name (optional): Specific Cohere model name.
            api_key (optional): Cohere key.
            connector_body (optional): The connector request body.
            connector_secret_name (optional): The connector secret name.

        Returns:
            bool: True if connector creation successful, False otherwise.
        """
        # Set trusted connector endpoints for Cohere
        trusted_endpoint = "^https://api\\.cohere\\.ai/.*$"
        self.set_trusted_endpoint(helper, trusted_endpoint)

        # Prompt to choose model
        model_type = self.get_model_details("Cohere", self.service_type, model_name)

        # Prompt for API key
        cohere_api_key = self.set_api_key(api_key, "Cohere")

        # Get connector body
        connector_body = connector_body or self._get_connector_body(model_type)

        auth_value = f"Bearer {cohere_api_key}"
        connector_body = json.loads(
            json.dumps(connector_body).replace("${auth}", auth_value)
        )

        if self.service_type == self.AMAZON_OPENSEARCH_SERVICE:
            # Create connector role and secret name
            connector_role_name, create_connector_role_name = (
                self.create_connector_role(connector_role_prefix, "cohere")
            )
            connector_secret_name, secret_value = self.create_secret_name(
                connector_secret_name, "cohere", cohere_api_key
            )

            # Create connector
            print("\nCreating Cohere connector...")
            connector_id, connector_role_arn, connector_secret_arn = (
                helper.create_connector_with_secret(
                    connector_secret_name,
                    secret_value,
                    connector_role_name,
                    create_connector_role_name,
                    connector_body,
                    sleep_time_in_seconds=10,
                )
            )
        else:
            credential_value = cohere_api_key
            connector_body = json.loads(
                json.dumps(connector_body).replace("${credential}", credential_value)
            )

            # Create connector
            print("\nCreating Cohere connector...")
            connector_id = helper.create_connector(
                create_connector_role_name=None,
                body=connector_body,
            )

        if connector_id:
            print(
                f"{Fore.GREEN}\nSuccessfully created Cohere connector with ID: {connector_id}{Style.RESET_ALL}"
            )
            connector_output = helper.get_connector(connector_id)
            save_config_method(
                connector_id,
                connector_output,
                (
                    connector_role_name
                    if self.service_type == self.AMAZON_OPENSEARCH_SERVICE
                    else None
                ),
                (
                    connector_role_arn
                    if self.service_type == self.AMAZON_OPENSEARCH_SERVICE
                    else None
                ),
                (
                    connector_secret_name
                    if self.service_type == self.AMAZON_OPENSEARCH_SERVICE
                    else None
                ),
                (
                    connector_secret_arn
                    if self.service_type == self.AMAZON_OPENSEARCH_SERVICE
                    else None
                ),
            )
            return True
        else:
            print(f"{Fore.RED}Failed to create Cohere connector.{Style.RESET_ALL}")
            return False



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

from typing import List, Optional

from opensearchpy import OpenSearch
from opensearchpy.exceptions import NotFoundError

from opensearch_py_ml.ml_commons.ml_common_utils import ML_BASE_URI
from opensearch_py_ml.ml_commons.validators import (
    validate_create_model_group_parameters,
    validate_delete_model_group_parameters,
    validate_search_model_group_parameters,
    validate_update_model_group_parameters,
)


class ModelAccessControl:
    API_ENDPOINT = "model_groups"

    def __init__(self, os_client: OpenSearch):
        self.client = os_client

    def register_model_group(
        self,
        name: str,
        description: Optional[str] = None,
        access_mode: Optional[str] = "private",
        backend_roles: Optional[List[str]] = None,
        add_all_backend_roles: Optional[bool] = False,
    ):
        validate_create_model_group_parameters(
            name, description, access_mode, backend_roles, add_all_backend_roles
        )

        body = {"name": name, "add_all_backend_roles": add_all_backend_roles}
        if description:
            body["description"] = description
        if access_mode:
            body["access_mode"] = access_mode
        if backend_roles:
            body["backend_roles"] = backend_roles

        return self.client.transport.perform_request(
            method="POST", url=f"{ML_BASE_URI}/{self.API_ENDPOINT}/_register", body=body
        )

    def update_model_group(
        self,
        update_query: dict,
        model_group_id: Optional[str] = None,
    ):
        validate_update_model_group_parameters(update_query, model_group_id)
        return self.client.transport.perform_request(
            method="PUT",
            url=f"{ML_BASE_URI}/{self.API_ENDPOINT}/{model_group_id}",
            body=update_query,
        )

    def search_model_group(self, query: dict):
        validate_search_model_group_parameters(query)
        return self.client.transport.perform_request(
            method="GET", url=f"{ML_BASE_URI}/{self.API_ENDPOINT}/_search", body=query
        )

    def search_model_group_by_name(
        self,
        model_group_name: str,
        _source: Optional[List] = None,
        size: Optional[int] = 1,
    ):
        query = {"query": {"match": {"name": model_group_name}}, "size": size}
        if _source:
            query["_source"] = _source
        return self.search_model_group(query)

    def get_model_group_id_by_name(self, model_group_name: str):
        try:
            res = self.search_model_group_by_name(model_group_name)
            if res["hits"]["hits"]:
                return res["hits"]["hits"][0]["_id"]
            else:
                raise NotFoundError
        except NotFoundError:
            print(f"No model group found with name:{model_group_name}")
            return None
        except Exception as ex:
            print(f"Error in get_model_group_id_by_name: {ex}")
            return None

    def delete_model_group(self, model_group_id: str):
        validate_delete_model_group_parameters(model_group_id)
        return self.client.transport.perform_request(
            method="DELETE", url=f"{ML_BASE_URI}/{self.API_ENDPOINT}/{model_group_id}"
        )

    def delete_model_group_by_name(self, model_group_name: str):
        model_group_id = self.get_model_group_id_by_name(model_group_name)
        if model_group_id is None:
            raise NotFoundError(f"Model group {model_group_name} not found")
        return self.delete_model_group(model_group_id=model_group_id)



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

"""
Main CLI script for connector creation
"""

import argparse
import logging
import os
import sys

from colorama import Fore, Style, init
from rich.console import Console

from opensearch_py_ml.ml_commons.cli.connector_manager import ConnectorManager
from opensearch_py_ml.ml_commons.cli.ml_setup import Setup
from opensearch_py_ml.ml_commons.cli.model_manager import ModelManager

# Initialize colorama for colored terminal output
init(autoreset=True)

# Initialize Rich console for enhanced CLI outputs
console = Console()

# Configure the logger for this module
logger = logging.getLogger(__name__)


class AllowDashActionConnector(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        """
        Enable argparse to correctly handle connector IDs that begin with a dash ('-')
        instead of treating them as argument separators.
        """
        if values is None:
            args = sys.argv
            try:
                connector_idx = args.index("--connectorId")
                if connector_idx + 1 < len(args):
                    value = args[connector_idx + 1]
                    setattr(namespace, self.dest, value)
                    sys.argv.remove("--connectorId")
                    sys.argv.remove(value)
                    return
            except ValueError:
                pass
            parser.error("--connectorId requires a value")
        else:
            setattr(namespace, self.dest, values)


class AllowDashActionModel(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        """
        Enable argparse to correctly handle model IDs that begin with a dash ('-')
        instead of treating them as argument separators.
        """
        if values is None:
            args = sys.argv
            try:
                model_idx = args.index("--modelId")
                if model_idx + 1 < len(args):
                    value = args[model_idx + 1]
                    setattr(namespace, self.dest, value)
                    sys.argv.remove("--modelId")
                    sys.argv.remove(value)
                    return
            except ValueError:
                pass
            parser.error("--modelId requires a value")
        else:
            setattr(namespace, self.dest, values)


def main():
    """
    Main function to handle opensearch-ml CLI commands
    """
    parser = argparse.ArgumentParser(
        prog="opensearch-ml",
        description="OpenSearch ML CLI",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog="""
Command Groups:
  connector
    create    Create a new connector

  model
    register  Register a new model
    predict   Predict a model

Examples:
  Initialize the setup:
    opensearch-ml setup

  Initialize the setup from a setup configuration file:
    opensearch-ml setup --path /file/setup_config.yml

  Create a connector:
    opensearch-ml connector create

  Create a connector from a connector configuration file:
    opensearch-ml connector create --path /file/connector_config.yml

  Register a model:
    opensearch-ml model register

  Register a model with a model name, description, and the connector ID
    opensearch-ml model register --connectorId 'connector123' --name 'Test model' --description 'This is a test model'

  Predict a model:
    opensearch-ml model predict

  Predict a model with a model ID and the request body:
    opensearch-ml model predict --modelId 'model123' --body '{"parameters": {"texts": ["hello world"]}}'
""",
    )
    # Create subparsers for different command groups
    subparsers = parser.add_subparsers(
        title="Available Commands", dest="command", metavar="command"
    )

    # Setup command
    setup = subparsers.add_parser(
        "setup", help="Initialize and configure OpenSearch setup and AWS credentials"
    )
    setup.add_argument(
        "--path", nargs="+", help="Path to the setup configuration file."
    )

    # Create the 'connector' command group
    connector_parser = subparsers.add_parser("connector", help="Manage ML connectors")

    # Create the 'model' command group
    model_parser = subparsers.add_parser("model", help="Manage ML models")

    # Create subcommands for 'connector'
    connector_subparsers = connector_parser.add_subparsers(
        title="Connector Commands", dest="subcommand", metavar="subcommand"
    )

    # Create command
    connector_create = connector_subparsers.add_parser(
        "create", help="Create a new connector"
    )
    connector_create.add_argument(
        "--path", nargs="+", help="Path to the connector configuration file."
    )

    # Create subcommands for 'model'
    model_subparsers = model_parser.add_subparsers(
        title="Model Commands", dest="subcommand", metavar="subcommand"
    )

    # Register command
    model_register = model_subparsers.add_parser(
        "register", help="Register a new model"
    )
    model_register.add_argument(
        "--connectorId",
        action=AllowDashActionConnector,
        help="The connector ID to register the model with.",
        metavar="CONNECTOR_ID",
        nargs="?",
    )
    model_register.add_argument(
        "--name",
        help="Name of the model to register.",
        metavar="MODEL_NAME",
    )
    model_register.add_argument(
        "--description",
        help="Description of the model to register.",
        metavar="MODEL_DESCRIPTION",
    )

    # Predict command
    model_predict = model_subparsers.add_parser("predict", help="Predict a model")
    model_predict.add_argument(
        "--modelId",
        action=AllowDashActionModel,
        help="ID of the model to predict.",
        metavar="MODEL_ID",
        nargs="?",
    )
    model_predict.add_argument(
        "--body", help="Body of the predict request", metavar="PREDICT_BODY"
    )

    args, unknown = parser.parse_known_args()
    config_path = None
    setup_config_path = None
    connector_manager = ConnectorManager()
    model_manager = ModelManager()

    if not args.command:
        parser.print_help()
        sys.exit(1)

    if args.command == "setup":
        setup = Setup()
        console.print("\n[bold blue]Starting connector setup...[/bold blue]")
        setup_config_path = args.path[0] if args.path else None
        config_path = setup.setup_command(config_path=setup_config_path)
        # Save the config path to a file for later use
        if config_path:
            config_dir = os.path.expanduser("~/.opensearch-ml")
            os.makedirs(config_dir, exist_ok=True)
            with open(os.path.join(config_dir, "config_path"), "w") as f:
                f.write(config_path)

    if args.command == "connector":
        if not args.subcommand:
            connector_parser.print_help()
            sys.exit(1)

        if args.subcommand == "create":
            console.print("\n[bold blue]Starting connector creation...[/bold blue]")
            connector_config_path = args.path[0] if args.path else None
            create_connector_result = connector_manager.initialize_create_connector(
                connector_config_path=connector_config_path
            )
            if create_connector_result:
                _, setup_config_path = create_connector_result
            # Save the setup config path after creation
            if setup_config_path:
                config_dir = os.path.expanduser("~/.opensearch-ml")
                os.makedirs(config_dir, exist_ok=True)
                with open(os.path.join(config_dir, "config_path"), "w") as f:
                    f.write(setup_config_path)

    if args.command == "model":
        if not args.subcommand:
            model_parser.print_help()
            sys.exit(1)

        if args.subcommand == "register":
            # Read the saved setup config path
            try:
                config_dir = os.path.expanduser("~/.opensearch-ml")
                with open(os.path.join(config_dir, "config_path"), "r") as f:
                    config_path = f.read().strip()
            except FileNotFoundError:
                logger.error(
                    f"{Fore.RED}No setup configuration found. Please run setup first.{Style.RESET_ALL}"
                )
                sys.exit(1)

            console.print("\n[bold blue]Starting model registration...[/bold blue]")
            connector_id = getattr(args, "connectorId", None)
            model_name = args.name if args.name else None
            model_description = args.description if args.description else None
            model_manager.initialize_register_model(
                config_path,
                connector_id=connector_id,
                model_name=model_name,
                model_description=model_description,
            )
        elif args.subcommand == "predict":
            # Read the saved setup config path
            try:
                config_dir = os.path.expanduser("~/.opensearch-ml")
                with open(os.path.join(config_dir, "config_path"), "r") as f:
                    config_path = f.read().strip()
            except FileNotFoundError:
                logger.error(
                    f"{Fore.RED}No setup configuration found. Please run setup first.{Style.RESET_ALL}"
                )
                sys.exit(1)
            console.print("\n[bold blue]Starting model prediction...[/bold blue]")
            model_id = args.modelId if args.modelId else None
            body = args.body if args.body else None
            model_manager.initialize_predict_model(
                config_path, model_id=model_id, body=body
            )


if __name__ == "__main__":
    main()
